/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Analyzer_instances, _Analyzer_packageJsonReader, _Analyzer_placeholders, _Analyzer_ongoingWorkPromises, _Analyzer_relevantConfigFilePaths, _Analyzer_getDiagnostics, _Analyzer_waitForAnalysisToComplete, _Analyzer_getPlaceholder, _Analyzer_upgradePlaceholder, _Analyzer_processDependencies, _Analyzer_processFiles, _Analyzer_processOutput, _Analyzer_processClean, _Analyzer_processPackageLocks, _Analyzer_checkForCyclesAndSortDependencies, _Analyzer_markAsInvalid, _Analyzer_resolveDependency, _Analyzer_resolveCrossPackageDependency;
import * as pathlib from 'path';
import { CachingPackageJsonReader, } from './util/package-json-reader.js';
import { scriptReferenceToString } from './script.js';
import { findNodeAtLocation } from './util/ast.js';
/**
 * Analyzes and validates a script along with all of its transitive
 * dependencies, producing a build graph that is ready to be executed.
 */
export class Analyzer {
    constructor(filesystem) {
        _Analyzer_instances.add(this);
        _Analyzer_packageJsonReader.set(this, void 0);
        _Analyzer_placeholders.set(this, new Map());
        _Analyzer_ongoingWorkPromises.set(this, []);
        _Analyzer_relevantConfigFilePaths.set(this, new Set());
        __classPrivateFieldSet(this, _Analyzer_packageJsonReader, new CachingPackageJsonReader(filesystem), "f");
    }
    /**
     * Analyze every script in each given file and return all diagnostics found.
     */
    async analyzeFiles(files) {
        await Promise.all(files.map(async (f) => {
            const packageDir = pathlib.dirname(f);
            const fileResult = await this.getPackageJson(packageDir);
            if (!fileResult.ok) {
                return; // will get this error below.
            }
            for (const script of fileResult.value.scripts) {
                // This starts analysis of each of the scripts in our root files.
                __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getPlaceholder).call(this, { name: script.name, packageDir });
            }
        }));
        await __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_waitForAnalysisToComplete).call(this);
        // Check for cycles.
        for (const info of __classPrivateFieldGet(this, _Analyzer_placeholders, "f").values()) {
            if (info.placeholder.state === 'unvalidated') {
                continue;
            }
            // We don't care about the result, if there's a cycle error it'll
            // be added to the scripts' diagnostics.
            __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_checkForCyclesAndSortDependencies).call(this, info.placeholder, new Set());
        }
        return __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getDiagnostics).call(this);
    }
    /**
     * Load the Wireit configuration from the `package.json` corresponding to the
     * given script, repeat for all transitive dependencies, and return a build
     * graph that is ready to be executed.
     *
     * Returns a Failure if the given script or any of its transitive
     * dependencies don't exist, are configured in an invalid way, or if there is
     * a cycle in the dependency graph.
     */
    async analyze(root, extraArgs) {
        // We do 2 walks through the dependency graph:
        //
        // 1. A non-deterministically ordered walk, where we traverse edges as soon
        //    as they are known, to maximize the parallelism of package.json file
        //    read operations.
        //
        // 2. A depth-first walk to detect cycles.
        //
        // We can't check for cycles in the 1st walk because its non-deterministic
        // traversal order means that we could miss certain cycle configurations.
        // Plus by doing a separate DFS walk, we'll always return the exact same
        // trail in the error message for any given graph, instead of an arbitrary
        // one.
        //
        // The way we avoid getting stuck in cycles during the 1st walk is by
        // allocating an initial placeholder object for each script, and caching it
        // by package + name. Then, instead of blocking each script on its
        // dependencies (which would lead to a promise cycle if there was a cycle in
        // the configuration), we wait for all placeholders to upgrade to full
        // configs asynchronously.
        const rootPlaceholder = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getPlaceholder).call(this, root);
        // Note we can't use Promise.all here, because new promises can be added to
        // the promises array as long as any promise is pending.
        await __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_waitForAnalysisToComplete).call(this);
        {
            const errors = await __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getDiagnostics).call(this);
            if (errors.size > 0) {
                return {
                    config: { ok: false, error: [...errors] },
                    relevantConfigFilePaths: __classPrivateFieldGet(this, _Analyzer_relevantConfigFilePaths, "f"),
                };
            }
        }
        // We can safely assume all placeholders have now been upgraded to full
        // configs.
        const rootConfig = rootPlaceholder.placeholder;
        if (rootConfig.state === 'unvalidated') {
            throw new Error(`Internal error: script ${root.name} in ${root.packageDir} is still unvalidated but had no failures`);
        }
        const cycleResult = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_checkForCyclesAndSortDependencies).call(this, rootConfig, new Set());
        if (!cycleResult.ok) {
            return {
                config: { ok: false, error: [cycleResult.error.dependencyFailure] },
                relevantConfigFilePaths: __classPrivateFieldGet(this, _Analyzer_relevantConfigFilePaths, "f"),
            };
        }
        const validRootConfig = cycleResult.value;
        validRootConfig.extraArgs = extraArgs;
        return {
            config: { ok: true, value: validRootConfig },
            relevantConfigFilePaths: __classPrivateFieldGet(this, _Analyzer_relevantConfigFilePaths, "f"),
        };
    }
    async analyzeIgnoringErrors(scriptReference) {
        await this.analyze(scriptReference, []);
        return __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getPlaceholder).call(this, scriptReference).placeholder;
    }
    async getPackageJson(packageDir) {
        __classPrivateFieldGet(this, _Analyzer_relevantConfigFilePaths, "f").add(pathlib.join(packageDir, 'package.json'));
        return __classPrivateFieldGet(this, _Analyzer_packageJsonReader, "f").read(packageDir);
    }
}
_Analyzer_packageJsonReader = new WeakMap(), _Analyzer_placeholders = new WeakMap(), _Analyzer_ongoingWorkPromises = new WeakMap(), _Analyzer_relevantConfigFilePaths = new WeakMap(), _Analyzer_instances = new WeakSet(), _Analyzer_getDiagnostics = async function _Analyzer_getDiagnostics() {
    const failures = new Set();
    for await (const failure of __classPrivateFieldGet(this, _Analyzer_packageJsonReader, "f").getFailures()) {
        failures.add(failure);
    }
    for (const info of __classPrivateFieldGet(this, _Analyzer_placeholders, "f").values()) {
        for (const failure of info.placeholder.failures) {
            failures.add(failure);
        }
    }
    for (const failure of failures) {
        const supercedes = failure
            .supercedes;
        if (supercedes != null) {
            failures.delete(supercedes);
        }
    }
    return failures;
}, _Analyzer_waitForAnalysisToComplete = async function _Analyzer_waitForAnalysisToComplete() {
    while (__classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").length > 0) {
        const promise = __classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f")[__classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").length - 1];
        await promise;
        // Need to be careful here. The contract of this method is that it does
        // not return until all pending analysis work is completed.
        // If there are multiple concurrent callers to this method, we want to
        // make sure that none of them hide any of the pending work from each
        // other by removing a promise from the array before it has settled.
        // So we first await the promise, and then remove it from the array if
        // it's still the final element.
        // It might not be the final element because another caller removed it,
        // or because more work was added onto the end of the array. Either
        // case is fine.
        if (promise ===
            __classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f")[__classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").length - 1]) {
            void __classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").pop();
        }
    }
}, _Analyzer_getPlaceholder = function _Analyzer_getPlaceholder(reference) {
    const scriptKey = scriptReferenceToString(reference);
    let placeholderInfo = __classPrivateFieldGet(this, _Analyzer_placeholders, "f").get(scriptKey);
    if (placeholderInfo === undefined) {
        const placeholder = {
            ...reference,
            state: 'unvalidated',
            failures: [],
        };
        placeholderInfo = {
            placeholder: placeholder,
            upgradeComplete: __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_upgradePlaceholder).call(this, placeholder),
        };
        __classPrivateFieldGet(this, _Analyzer_placeholders, "f").set(scriptKey, placeholderInfo);
        __classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").push(placeholderInfo.upgradeComplete);
    }
    return placeholderInfo;
}, _Analyzer_upgradePlaceholder = 
/**
 * In-place upgrade the given placeholder script configuration object to a
 * full configuration, by reading its package.json file.
 *
 * Note this method does not block on the script's dependencies being
 * upgraded; dependencies are upgraded asynchronously.
 */
async function _Analyzer_upgradePlaceholder(placeholder) {
    const packageJsonResult = await this.getPackageJson(placeholder.packageDir);
    if (!packageJsonResult.ok) {
        placeholder.failures.push(packageJsonResult.error);
        return undefined;
    }
    const packageJson = packageJsonResult.value;
    placeholder.failures.push(...packageJson.failures);
    const syntaxInfo = packageJson.getScriptInfo(placeholder.name);
    if (syntaxInfo === undefined || syntaxInfo.scriptNode === undefined) {
        let node;
        let reason;
        if (syntaxInfo?.wireitConfigNode?.name != null) {
            node = syntaxInfo.wireitConfigNode.name;
            reason = 'wireit-config-but-no-script';
        }
        else {
            node ?? (node = packageJson.scriptsSection?.name);
            reason = 'script-not-found';
        }
        const range = node
            ? { offset: node.offset, length: node.length }
            : { offset: 0, length: 0 };
        placeholder.failures.push({
            type: 'failure',
            reason,
            script: placeholder,
            diagnostic: {
                severity: 'error',
                message: `Script "${placeholder.name}" not found in the scripts section of this package.json.`,
                location: { file: packageJson.jsonFile, range },
            },
        });
        return undefined;
    }
    const scriptCommand = syntaxInfo.scriptNode;
    const wireitConfig = syntaxInfo.wireitConfigNode;
    if (wireitConfig !== undefined &&
        scriptCommand.value !== 'wireit' &&
        scriptCommand.value !== 'yarn run -TB wireit') {
        const configName = wireitConfig.name;
        placeholder.failures.push({
            type: 'failure',
            reason: 'script-not-wireit',
            script: placeholder,
            diagnostic: {
                message: `This command should just be "wireit", as this script is configured in the wireit section.`,
                severity: 'warning',
                location: {
                    file: packageJson.jsonFile,
                    range: {
                        length: scriptCommand.length,
                        offset: scriptCommand.offset,
                    },
                },
                supplementalLocations: [
                    {
                        message: `The wireit config is here.`,
                        location: {
                            file: packageJson.jsonFile,
                            range: {
                                length: configName.length,
                                offset: configName.offset,
                            },
                        },
                    },
                ],
            },
        });
    }
    if (wireitConfig === undefined && scriptCommand.value === 'wireit') {
        placeholder.failures.push({
            type: 'failure',
            reason: 'invalid-config-syntax',
            script: placeholder,
            diagnostic: {
                severity: 'error',
                message: `This script is configured to run wireit but it has no config in the wireit section of this package.json file`,
                location: {
                    file: packageJson.jsonFile,
                    range: {
                        length: scriptCommand.length,
                        offset: scriptCommand.offset,
                    },
                },
            },
        });
    }
    const { dependencies, encounteredError: dependenciesErrored } = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_processDependencies).call(this, placeholder, packageJson, syntaxInfo);
    let command;
    let commandError = false;
    if (wireitConfig === undefined) {
        const result = failUnlessNonBlankString(scriptCommand, packageJson.jsonFile);
        if (result.ok) {
            command = result.value;
        }
        else {
            commandError = true;
            placeholder.failures.push(result.error);
        }
    }
    else {
        const commandAst = findNodeAtLocation(wireitConfig, ['command']);
        if (commandAst !== undefined) {
            const result = failUnlessNonBlankString(commandAst, packageJson.jsonFile);
            if (result.ok) {
                command = result.value;
            }
            else {
                commandError = true;
                placeholder.failures.push(result.error);
            }
        }
    }
    const files = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_processFiles).call(this, placeholder, packageJson, syntaxInfo);
    if (wireitConfig !== undefined &&
        dependencies.length === 0 &&
        !dependenciesErrored &&
        command === undefined &&
        !commandError &&
        (files === undefined || files.values.length === 0)) {
        placeholder.failures.push({
            type: 'failure',
            reason: 'invalid-config-syntax',
            script: placeholder,
            diagnostic: {
                severity: 'error',
                message: `A wireit config must set at least one of "command", "dependencies", or "files". Otherwise there is nothing for wireit to do.`,
                location: {
                    file: packageJson.jsonFile,
                    range: {
                        length: wireitConfig.name.length,
                        offset: wireitConfig.name.offset,
                    },
                },
            },
        });
    }
    const output = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_processOutput).call(this, placeholder, packageJson, syntaxInfo, command);
    const clean = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_processClean).call(this, placeholder, packageJson, syntaxInfo);
    __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_processPackageLocks).call(this, placeholder, packageJson, syntaxInfo, files);
    // It's important to in-place update the placeholder object, instead of
    // creating a new object, because other configs may be referencing this
    // exact object in their dependencies.
    const remainingConfig = {
        ...placeholder,
        state: 'locally-valid',
        failures: placeholder.failures,
        command,
        extraArgs: undefined,
        dependencies,
        files,
        output,
        clean: clean ?? true,
        scriptAstNode: scriptCommand,
        configAstNode: wireitConfig,
        declaringFile: packageJson.jsonFile,
    };
    Object.assign(placeholder, remainingConfig);
}, _Analyzer_processDependencies = function _Analyzer_processDependencies(placeholder, packageJson, scriptInfo) {
    const dependencies = [];
    const dependenciesAst = scriptInfo.wireitConfigNode &&
        findNodeAtLocation(scriptInfo.wireitConfigNode, ['dependencies']);
    let encounteredError = false;
    if (dependenciesAst == null) {
        return { dependencies, encounteredError };
    }
    const result = failUnlessArray(dependenciesAst, packageJson.jsonFile);
    if (!result.ok) {
        encounteredError = true;
        placeholder.failures.push(result.error);
        return { dependencies, encounteredError };
    }
    // Error if the same dependency is declared multiple times. Duplicate
    // dependencies aren't necessarily a serious problem (since we already
    // prevent double-analysis here, and double-analysis in the Executor), but
    // they may indicate that the user has made a mistake (e.g. maybe they
    // meant a different dependency).
    const uniqueDependencies = new Map();
    const children = dependenciesAst.children ?? [];
    for (let i = 0; i < children.length; i++) {
        const maybeUnresolved = children[i];
        const stringResult = failUnlessNonBlankString(maybeUnresolved, packageJson.jsonFile);
        if (!stringResult.ok) {
            encounteredError = true;
            placeholder.failures.push(stringResult.error);
            continue;
        }
        const unresolved = stringResult.value;
        const result = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_resolveDependency).call(this, unresolved, placeholder, packageJson.jsonFile);
        if (!result.ok) {
            encounteredError = true;
            placeholder.failures.push(result.error);
            continue;
        }
        for (const resolved of result.value) {
            const uniqueKey = scriptReferenceToString(resolved);
            const duplicate = uniqueDependencies.get(uniqueKey);
            if (duplicate !== undefined) {
                encounteredError = true;
                placeholder.failures.push({
                    type: 'failure',
                    reason: 'duplicate-dependency',
                    script: placeholder,
                    dependency: resolved,
                    diagnostic: {
                        severity: 'error',
                        message: `This dependency is listed multiple times`,
                        location: {
                            file: packageJson.jsonFile,
                            range: {
                                offset: unresolved.offset,
                                length: unresolved.length,
                            },
                        },
                        supplementalLocations: [
                            {
                                message: `The dependency was first listed here.`,
                                location: {
                                    file: packageJson.jsonFile,
                                    range: {
                                        offset: duplicate.offset,
                                        length: duplicate.length,
                                    },
                                },
                            },
                        ],
                    },
                });
            }
            uniqueDependencies.set(uniqueKey, unresolved);
            const placeHolderInfo = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_getPlaceholder).call(this, resolved);
            dependencies.push({
                astNode: unresolved,
                config: placeHolderInfo.placeholder,
            });
            __classPrivateFieldGet(this, _Analyzer_ongoingWorkPromises, "f").push((async () => {
                await placeHolderInfo.upgradeComplete;
                const failures = placeHolderInfo.placeholder.failures;
                for (const failure of failures) {
                    if (failure.reason === 'script-not-found') {
                        const hasColon = unresolved.value.includes(':');
                        let offset;
                        let length;
                        if (!hasColon ||
                            resolved.packageDir === placeholder.packageDir) {
                            offset = unresolved.offset;
                            length = unresolved.length;
                        }
                        else {
                            // Skip past the colon
                            const colonOffsetInString = packageJson.jsonFile.contents
                                .slice(unresolved.offset)
                                .indexOf(':');
                            offset = unresolved.offset + colonOffsetInString + 1;
                            length = unresolved.length - colonOffsetInString - 2;
                        }
                        placeholder.failures.push({
                            type: 'failure',
                            reason: 'dependency-on-missing-script',
                            script: placeholder,
                            supercedes: failure,
                            diagnostic: {
                                severity: 'error',
                                message: `Cannot find script named ${JSON.stringify(resolved.name)} in package "${resolved.packageDir}"`,
                                location: {
                                    file: packageJson.jsonFile,
                                    range: { offset, length },
                                },
                            },
                        });
                    }
                    else if (failure.reason === 'missing-package-json') {
                        // Skip the opening "
                        const offset = unresolved.offset + 1;
                        // Take everything up to the first colon, but look in
                        // the original source, to avoid getting confused by escape
                        // sequences, which have a different length before and after
                        // encoding.
                        const length = packageJson.jsonFile.contents
                            .slice(offset)
                            .indexOf(':');
                        const range = { offset, length };
                        placeholder.failures.push({
                            type: 'failure',
                            reason: 'dependency-on-missing-package-json',
                            script: placeholder,
                            supercedes: failure,
                            diagnostic: {
                                severity: 'error',
                                message: `package.json file missing: "${pathlib.join(resolved.packageDir, 'package.json')}"`,
                                location: { file: packageJson.jsonFile, range },
                            },
                        });
                    }
                }
                return undefined;
            })());
        }
    }
    return { dependencies, encounteredError };
}, _Analyzer_processFiles = function _Analyzer_processFiles(placeholder, packageJson, syntaxInfo) {
    if (syntaxInfo.wireitConfigNode == null) {
        return;
    }
    const filesNode = findNodeAtLocation(syntaxInfo.wireitConfigNode, [
        'files',
    ]);
    if (filesNode === undefined) {
        return;
    }
    const values = [];
    const result = failUnlessArray(filesNode, packageJson.jsonFile);
    if (!result.ok) {
        placeholder.failures.push(result.error);
        return;
    }
    const children = filesNode.children ?? [];
    for (let i = 0; i < children.length; i++) {
        const file = children[i];
        const result = failUnlessNonBlankString(file, packageJson.jsonFile);
        if (!result.ok) {
            placeholder.failures.push(result.error);
            continue;
        }
        values.push(result.value.value);
    }
    return { node: filesNode, values };
}, _Analyzer_processOutput = function _Analyzer_processOutput(placeholder, packageJson, syntaxInfo, command) {
    if (syntaxInfo.wireitConfigNode == null) {
        return;
    }
    const outputNode = findNodeAtLocation(syntaxInfo.wireitConfigNode, [
        'output',
    ]);
    if (outputNode === undefined) {
        return;
    }
    if (command === undefined) {
        placeholder.failures.push({
            type: 'failure',
            reason: 'invalid-config-syntax',
            script: placeholder,
            diagnostic: {
                severity: 'error',
                message: `"output" can only be set if "command" is also set.`,
                location: {
                    file: packageJson.jsonFile,
                    range: {
                        // Highlight the whole `"output": []` part.
                        length: (outputNode.parent ?? outputNode).length,
                        offset: (outputNode.parent ?? outputNode).offset,
                    },
                },
            },
        });
    }
    const values = [];
    const result = failUnlessArray(outputNode, packageJson.jsonFile);
    if (!result.ok) {
        placeholder.failures.push(result.error);
        return;
    }
    const children = outputNode.children ?? [];
    for (let i = 0; i < children.length; i++) {
        const anOutput = children[i];
        const result = failUnlessNonBlankString(anOutput, packageJson.jsonFile);
        if (!result.ok) {
            placeholder.failures.push(result.error);
            continue;
        }
        values.push(result.value.value);
    }
    return { node: outputNode, values };
}, _Analyzer_processClean = function _Analyzer_processClean(placeholder, packageJson, syntaxInfo) {
    if (syntaxInfo.wireitConfigNode == null) {
        return;
    }
    const clean = findNodeAtLocation(syntaxInfo.wireitConfigNode, ['clean']);
    if (clean !== undefined &&
        clean.value !== true &&
        clean.value !== false &&
        clean.value !== 'if-file-deleted') {
        placeholder.failures.push({
            type: 'failure',
            reason: 'invalid-config-syntax',
            script: placeholder,
            diagnostic: {
                severity: 'error',
                message: `The "clean" property must be either true, false, or "if-file-deleted".`,
                location: {
                    file: packageJson.jsonFile,
                    range: { length: clean.length, offset: clean.offset },
                },
            },
        });
        // We shouldn't execute if there's failures, but just in case, this is
        // likely the safest option.
        return false;
    }
    return clean?.value;
}, _Analyzer_processPackageLocks = function _Analyzer_processPackageLocks(placeholder, packageJson, syntaxInfo, files) {
    if (syntaxInfo.wireitConfigNode == null) {
        return;
    }
    const packageLocksNode = findNodeAtLocation(syntaxInfo.wireitConfigNode, [
        'packageLocks',
    ]);
    let packageLocks;
    if (packageLocksNode !== undefined) {
        const result = failUnlessArray(packageLocksNode, packageJson.jsonFile);
        if (!result.ok) {
            placeholder.failures.push(result.error);
        }
        else {
            packageLocks = { node: packageLocksNode, values: [] };
            const children = packageLocksNode.children ?? [];
            for (let i = 0; i < children.length; i++) {
                const maybeFilename = children[i];
                const result = failUnlessNonBlankString(maybeFilename, packageJson.jsonFile);
                if (!result.ok) {
                    placeholder.failures.push(result.error);
                    continue;
                }
                const filename = result.value;
                if (filename.value !== pathlib.basename(filename.value)) {
                    placeholder.failures.push({
                        type: 'failure',
                        reason: 'invalid-config-syntax',
                        script: placeholder,
                        diagnostic: {
                            severity: 'error',
                            message: `A package lock must be a filename, not a path`,
                            location: {
                                file: packageJson.jsonFile,
                                range: { length: filename.length, offset: filename.offset },
                            },
                        },
                    });
                    continue;
                }
                packageLocks.values.push(filename.value);
            }
        }
    }
    if (
    // There's no reason to check package locks when "files" is undefined,
    // because scripts will always run in that case anyway.
    files !== undefined &&
        // An explicitly empty "packageLocks" array disables package lock checking
        // entirely.
        packageLocks?.values.length !== 0) {
        const lockfileNames = packageLocks?.values ?? ['package-lock.json'];
        // Generate "package-lock.json", "../package-lock.json",
        // "../../package-lock.json" etc. all the way up to the root of the
        // filesystem, because that's how Node package resolution works.
        const depth = placeholder.packageDir.split(pathlib.sep).length;
        for (let i = 0; i < depth; i++) {
            // Glob patterns are specified with forward-slash delimiters, even on
            // Windows.
            const prefix = Array(i + 1).join('../');
            for (const lockfileName of lockfileNames) {
                files.values.push(prefix + lockfileName);
            }
        }
    }
}, _Analyzer_checkForCyclesAndSortDependencies = function _Analyzer_checkForCyclesAndSortDependencies(config, trail) {
    if (config.state === 'valid') {
        // Already validated.
        return { ok: true, value: config };
    }
    else if (config.state === 'invalid') {
        return { ok: false, error: config };
    }
    let dependencyStillUnvalidated = undefined;
    const trailKey = scriptReferenceToString(config);
    const supplementalLocations = [];
    if (trail.has(trailKey)) {
        // Found a cycle.
        let cycleStart = 0;
        // Trail is in graph traversal order because JavaScript Map iteration
        // order matches insertion order.
        let i = 0;
        for (const visitedKey of trail) {
            if (visitedKey === trailKey) {
                cycleStart = i;
            }
            i++;
        }
        const trailArray = [...trail].map((key) => {
            const placeholderInfo = __classPrivateFieldGet(this, _Analyzer_placeholders, "f").get(key);
            if (placeholderInfo == null) {
                throw new Error(`Internal error: placeholder not found for ${key} during cycle detection`);
            }
            return placeholderInfo.placeholder;
        });
        trailArray.push(config);
        const cycleEnd = trailArray.length - 1;
        for (let i = cycleStart; i < cycleEnd; i++) {
            const current = trailArray[i];
            const next = trailArray[i + 1];
            if (current.state === 'unvalidated') {
                dependencyStillUnvalidated = current;
                continue;
            }
            const nextNode = current.dependencies.find((dep) => dep.config === next);
            // Use the actual value in the array, because this could refer to
            // a script in another package.
            const nextName = nextNode?.astNode?.value ?? next?.name ?? trailArray[cycleStart].name;
            const message = next === trailArray[cycleStart]
                ? `${JSON.stringify(current.name)} points back to ${JSON.stringify(nextName)}`
                : `${JSON.stringify(current.name)} points to ${JSON.stringify(nextName)}`;
            const culpritNode = 
            // This should always be present
            nextNode?.astNode ??
                // But failing that, fall back to the best node we have.
                current.configAstNode?.name ??
                current.scriptAstNode?.name;
            supplementalLocations.push({
                message,
                location: {
                    file: current.declaringFile,
                    range: {
                        offset: culpritNode.offset,
                        length: culpritNode.length,
                    },
                },
            });
        }
        const diagnostic = {
            severity: 'error',
            message: `Cycle detected in dependencies of ${JSON.stringify(config.name)}.`,
            location: {
                file: config.declaringFile,
                range: {
                    length: config.configAstNode?.name.length ??
                        config.scriptAstNode?.name.length,
                    offset: config.configAstNode?.name.offset ??
                        config.scriptAstNode?.name.length,
                },
            },
            supplementalLocations,
        };
        const failure = {
            type: 'failure',
            reason: 'cycle',
            script: config,
            diagnostic,
        };
        return { ok: false, error: __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_markAsInvalid).call(this, config, failure) };
    }
    if (config.dependencies != null && config.dependencies.length > 0) {
        // Sorting means that if the user re-orders the same set of dependencies,
        // the trail we take in this walk remains the same, so any cycle error
        // message we might throw will have the same trail, too. This also helps
        // make the caching keys that we'll be generating in the later execution
        // step insensitive to dependency order as well.
        config.dependencies.sort((a, b) => {
            if (a.config.packageDir !== b.config.packageDir) {
                return a.config.packageDir.localeCompare(b.config.packageDir);
            }
            return a.config.name.localeCompare(b.config.name);
        });
        trail.add(trailKey);
        for (const dependency of config.dependencies) {
            if (dependency.config.state === 'unvalidated') {
                dependencyStillUnvalidated = dependency.config;
                continue;
            }
            const result = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_checkForCyclesAndSortDependencies).call(this, dependency.config, trail);
            if (!result.ok) {
                return {
                    ok: false,
                    error: __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_markAsInvalid).call(this, config, result.error.dependencyFailure),
                };
            }
        }
        trail.delete(trailKey);
    }
    if (dependencyStillUnvalidated != null) {
        // At least one of our dependencies was unvalidated, likely because it
        // had a syntax error or was missing necessary information. Therefore
        // we can't transition to valid either.
        const failure = {
            type: 'failure',
            reason: 'dependency-invalid',
            script: config,
            dependency: dependencyStillUnvalidated,
        };
        return { ok: false, error: __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_markAsInvalid).call(this, config, failure) };
    }
    {
        const validConfig = {
            ...config,
            extraArgs: undefined,
            state: 'valid',
            dependencies: config.dependencies,
        };
        // We want to keep the original reference, but get type checking that
        // the only difference between a ScriptConfig and a
        // LocallyValidScriptConfig is that the state is 'valid' and the
        // dependencies are also valid, which we confirmed above.
        Object.assign(config, validConfig);
    }
    return { ok: true, value: config };
}, _Analyzer_markAsInvalid = function _Analyzer_markAsInvalid(config, failure) {
    const invalidConfig = {
        ...config,
        state: 'invalid',
        dependencyFailure: failure,
    };
    Object.assign(config, invalidConfig);
    config.failures.push(failure);
    return config;
}, _Analyzer_resolveDependency = function _Analyzer_resolveDependency(dependency, context, referencingFile) {
    // TODO(aomarks) Implement $WORKSPACES syntax.
    if (dependency.value.startsWith('.')) {
        // TODO(aomarks) It is technically valid for an npm script to start with a
        // ".". We should support that edge case with backslash escaping.
        const result = __classPrivateFieldGet(this, _Analyzer_instances, "m", _Analyzer_resolveCrossPackageDependency).call(this, dependency, context, referencingFile);
        if (!result.ok) {
            return result;
        }
        return { ok: true, value: [result.value] };
    }
    return {
        ok: true,
        value: [{ packageDir: context.packageDir, name: dependency.value }],
    };
}, _Analyzer_resolveCrossPackageDependency = function _Analyzer_resolveCrossPackageDependency(dependency, context, referencingFile) {
    // TODO(aomarks) On some file systems, it is valid to have a ":" in a file
    // path. We should support that edge case with backslash escaping.
    const firstColonIdx = dependency.value.indexOf(':');
    if (firstColonIdx === -1) {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: context,
                diagnostic: {
                    severity: 'error',
                    message: `Cross-package dependency must use syntax ` +
                        `"<relative-path>:<script-name>", ` +
                        `but there's no ":" character in "${dependency.value}".`,
                    location: {
                        file: referencingFile,
                        range: { offset: dependency.offset, length: dependency.length },
                    },
                },
            },
        };
    }
    const scriptName = dependency.value.slice(firstColonIdx + 1);
    if (!scriptName) {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: context,
                diagnostic: {
                    severity: 'error',
                    message: `Cross-package dependency must use syntax ` +
                        `"<relative-path>:<script-name>", ` +
                        `but there's no script name in "${dependency.value}".`,
                    location: {
                        file: referencingFile,
                        range: { offset: dependency.offset, length: dependency.length },
                    },
                },
            },
        };
    }
    const relativePackageDir = dependency.value.slice(0, firstColonIdx);
    const absolutePackageDir = pathlib.resolve(context.packageDir, relativePackageDir);
    if (absolutePackageDir === context.packageDir) {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: context,
                diagnostic: {
                    severity: 'error',
                    message: `Cross-package dependency "${dependency.value}" ` +
                        `resolved to the same package.`,
                    location: {
                        file: referencingFile,
                        range: { offset: dependency.offset, length: dependency.length },
                    },
                },
            },
        };
    }
    return {
        ok: true,
        value: { packageDir: absolutePackageDir, name: scriptName },
    };
};
export function failUnlessNonBlankString(astNode, file) {
    if (astNode.type !== 'string') {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: { packageDir: pathlib.dirname(file.path) },
                diagnostic: {
                    severity: 'error',
                    message: `Expected a string, but was ${astNode.type}.`,
                    location: {
                        file,
                        range: {
                            offset: astNode.offset,
                            length: astNode.length,
                        },
                    },
                },
            },
        };
    }
    if (astNode.value.match(/^\s*$/)) {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: { packageDir: pathlib.dirname(file.path) },
                diagnostic: {
                    severity: 'error',
                    message: `Expected this field to be nonempty`,
                    location: {
                        file,
                        range: {
                            offset: astNode.offset,
                            length: astNode.length,
                        },
                    },
                },
            },
        };
    }
    return { ok: true, value: astNode };
}
/**
 * Return a failing result if the given value is not an Array.
 */
const failUnlessArray = (astNode, file) => {
    if (astNode.type !== 'array') {
        return {
            ok: false,
            error: {
                type: 'failure',
                reason: 'invalid-config-syntax',
                script: { packageDir: pathlib.dirname(file.path) },
                diagnostic: {
                    severity: 'error',
                    message: `Expected an array, but was ${astNode.type}.`,
                    location: {
                        file: file,
                        range: {
                            offset: astNode.offset,
                            length: astNode.length,
                        },
                    },
                },
            },
        };
    }
    return { ok: true, value: undefined };
};
/**
 * Return a failed result if the given value is not an object literal ({...}).
 */
export const failUnlessJsonObject = (astNode, file) => {
    if (astNode.type !== 'object') {
        return {
            type: 'failure',
            reason: 'invalid-config-syntax',
            script: { packageDir: pathlib.dirname(file.path) },
            diagnostic: {
                severity: 'error',
                message: `Expected an object, but was ${astNode.type}.`,
                location: {
                    file: file,
                    range: {
                        offset: astNode.offset,
                        length: astNode.length,
                    },
                },
            },
        };
    }
};
//# sourceMappingURL=analyzer.js.map