/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var _LocalCache_instances, _LocalCache_getCacheDir, _LocalCacheHit_source, _LocalCacheHit_destination;
import * as fs from 'fs/promises';
import * as pathlib from 'path';
import { createHash } from 'crypto';
import { getScriptDataDir } from '../util/script-data-dir.js';
import { copyEntries } from '../util/copy.js';
import { glob } from '../util/glob.js';
/**
 * Caches script output to each package's
 * ".wireit/<script-name-hex>/cache/<cache-key-sha256-hex>" folder.
 */
export class LocalCache {
    constructor() {
        _LocalCache_instances.add(this);
    }
    async get(script, fingerprint) {
        const cacheDir = __classPrivateFieldGet(this, _LocalCache_instances, "m", _LocalCache_getCacheDir).call(this, script, fingerprint);
        try {
            await fs.access(cacheDir);
        }
        catch (error) {
            if (error.code === 'ENOENT') {
                return;
            }
            throw error;
        }
        return new LocalCacheHit(cacheDir, script.packageDir);
    }
    async set(script, fingerprint, absoluteFiles) {
        // TODO(aomarks) A script's cache directory currently just grows forever.
        // We'll have the "clean" command to help with manual cleanup, but we'll
        // almost certainly want an automated way to limit the size of the cache
        // directory (e.g. LRU capped to some number of entries).
        // https://github.com/google/wireit/issues/71
        const absCacheDir = __classPrivateFieldGet(this, _LocalCache_instances, "m", _LocalCache_getCacheDir).call(this, script, fingerprint);
        // Note fs.mkdir returns the first created directory, or undefined if no
        // directory was created.
        const existed = (await fs.mkdir(absCacheDir, { recursive: true })) === undefined;
        if (existed) {
            // This is an unexpected error because the Executor should already have
            // checked for an existing cache hit.
            throw new Error(`Did not expect ${absCacheDir} to already exist.`);
        }
        await copyEntries(absoluteFiles, script.packageDir, absCacheDir);
        return true;
    }
}
_LocalCache_instances = new WeakSet(), _LocalCache_getCacheDir = function _LocalCache_getCacheDir(script, fingerprint) {
    return pathlib.join(getScriptDataDir(script), 'cache', createHash('sha256').update(fingerprint.string).digest('hex'));
};
class LocalCacheHit {
    constructor(source, destination) {
        /**
         * The folder where the cached output is stored. Assumed to exist.
         */
        _LocalCacheHit_source.set(this, void 0);
        /**
         * The folder where the cached output should be written when {@link apply} is
         * called.
         */
        _LocalCacheHit_destination.set(this, void 0);
        __classPrivateFieldSet(this, _LocalCacheHit_source, source, "f");
        __classPrivateFieldSet(this, _LocalCacheHit_destination, destination, "f");
    }
    async apply() {
        const entries = await glob(['**'], {
            cwd: __classPrivateFieldGet(this, _LocalCacheHit_source, "f"),
            followSymlinks: false,
            includeDirectories: true,
            expandDirectories: true,
            // Shouldn't ever happen, but would be really weird.
            throwIfOutsideCwd: true,
        });
        await copyEntries(entries, __classPrivateFieldGet(this, _LocalCacheHit_source, "f"), __classPrivateFieldGet(this, _LocalCacheHit_destination, "f"));
    }
}
_LocalCacheHit_source = new WeakMap(), _LocalCacheHit_destination = new WeakMap();
//# sourceMappingURL=local-cache.js.map