/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _DefaultLogger_instances, _DefaultLogger_rootPackageDir, _DefaultLogger_diagnosticPrinter, _DefaultLogger_label;
import * as pathlib from 'path';
import { unreachable } from '../util/unreachable.js';
import { DiagnosticPrinter } from '../error.js';
import { createRequire } from 'module';
const getWireitVersion = (() => {
    let version;
    return () => {
        if (version === undefined) {
            version = createRequire(import.meta.url)('../../package.json').version;
        }
        return version;
    };
})();
/**
 * Default {@link Logger} which logs to stdout and stderr.
 */
export class DefaultLogger {
    /**
     * @param rootPackage The npm package directory that the root script being
     * executed belongs to.
     */
    constructor(rootPackage) {
        _DefaultLogger_instances.add(this);
        _DefaultLogger_rootPackageDir.set(this, void 0);
        _DefaultLogger_diagnosticPrinter.set(this, void 0);
        __classPrivateFieldSet(this, _DefaultLogger_rootPackageDir, rootPackage, "f");
        __classPrivateFieldSet(this, _DefaultLogger_diagnosticPrinter, new DiagnosticPrinter(__classPrivateFieldGet(this, _DefaultLogger_rootPackageDir, "f")), "f");
    }
    log(event) {
        const type = event.type;
        const label = __classPrivateFieldGet(this, _DefaultLogger_instances, "m", _DefaultLogger_label).call(this, event.script);
        const prefix = label !== '' ? ` [${label}]` : '';
        switch (type) {
            default: {
                throw new Error(`Unknown event type: ${unreachable(type)}`);
            }
            case 'success': {
                const reason = event.reason;
                switch (reason) {
                    default: {
                        throw new Error(`Unknown success reason: ${unreachable(reason)}`);
                    }
                    case 'exit-zero': {
                        console.log(`✅${prefix} Executed successfully`);
                        break;
                    }
                    case 'no-command': {
                        console.log(`✅${prefix} No command to execute`);
                        break;
                    }
                    case 'fresh': {
                        console.log(`✅${prefix} Already fresh`);
                        break;
                    }
                    case 'cached': {
                        console.log(`✅${prefix} Restored from cache`);
                        break;
                    }
                }
                break;
            }
            case 'failure': {
                const reason = event.reason;
                switch (reason) {
                    default: {
                        throw new Error(`Unknown failure reason: ${unreachable(reason)}`);
                    }
                    case 'launched-incorrectly': {
                        console.error(`❌${prefix} wireit must be launched with "npm run" or a compatible command.`);
                        console.error(`    More info: ${event.detail}`);
                        break;
                    }
                    case 'missing-package-json': {
                        console.error(`❌${prefix} No package.json was found in ${event.script.packageDir}`);
                        break;
                    }
                    case 'invalid-json-syntax': {
                        for (const diagnostic of event.diagnostics) {
                            console.error(__classPrivateFieldGet(this, _DefaultLogger_diagnosticPrinter, "f").print(diagnostic));
                        }
                        break;
                    }
                    case 'no-scripts-in-package-json': {
                        console.error(`❌${prefix} No "scripts" section defined in package.json in ${event.script.packageDir}`);
                        break;
                    }
                    case 'script-not-found':
                    case 'wireit-config-but-no-script':
                    case 'duplicate-dependency':
                    case 'script-not-wireit':
                    case 'invalid-config-syntax':
                    case 'cycle':
                    case 'dependency-on-missing-package-json':
                    case 'dependency-on-missing-script': {
                        console.error(__classPrivateFieldGet(this, _DefaultLogger_diagnosticPrinter, "f").print(event.diagnostic));
                        break;
                    }
                    case 'invalid-usage': {
                        console.error(`❌${prefix} Invalid usage: ${event.message}`);
                        break;
                    }
                    case 'exit-non-zero': {
                        console.error(`❌${prefix} Failed with exit status ${event.status}`);
                        break;
                    }
                    case 'signal': {
                        console.error(`❌${prefix} Failed with signal ${event.signal}`);
                        break;
                    }
                    case 'spawn-error': {
                        console.error(`❌${prefix} Process spawn error: ${event.message}`);
                        break;
                    }
                    case 'start-cancelled': {
                        // The script never started. We don't really need to log this, it's
                        // fairly noisy. Maybe in a verbose mode.
                        break;
                    }
                    case 'killed': {
                        console.error(`💀${prefix} Killed`);
                        break;
                    }
                    case 'unknown-error-thrown': {
                        console.error(`❌${prefix} Internal error! Please file a bug at https://github.com/google/wireit/issues/new, mention this message, that you encountered it in wireit version ${getWireitVersion()}, and give information about your package.json files.\n    Unknown error thrown: ${String(event.error)}`);
                        const maybeError = event.error;
                        if (maybeError?.stack) {
                            console.error(maybeError.stack);
                        }
                        break;
                    }
                    case 'dependency-invalid': {
                        console.error(`❌${prefix} Depended, perhaps indirectly, on ${__classPrivateFieldGet(this, _DefaultLogger_instances, "m", _DefaultLogger_label).call(this, event.dependency)} which could not be validated. Please file a bug at https://github.com/google/wireit/issues/new, mention this message, that you encountered it in wireit version ${getWireitVersion()}, and give information about your package.json files.`);
                    }
                }
                break;
            }
            case 'output': {
                const stream = event.stream;
                switch (stream) {
                    default: {
                        throw new Error(`Unknown output stream: ${unreachable(stream)}`);
                    }
                    // TODO(aomarks) More advanced handling of output streams so that
                    // output isn't simply interweaved.
                    case 'stdout': {
                        process.stdout.write(event.data);
                        break;
                    }
                    case 'stderr': {
                        process.stderr.write(event.data);
                        break;
                    }
                }
                break;
            }
            case 'info': {
                const detail = event.detail;
                switch (detail) {
                    default: {
                        throw new Error(`Unknown info event detail: ${unreachable(detail)}`);
                    }
                    case 'running': {
                        console.log(`🏃${prefix} Running command "${event.script.command?.value ?? ''}"`);
                        break;
                    }
                    case 'locked': {
                        console.log(`💤${prefix} Waiting for another process which is already running this script.`);
                        break;
                    }
                    case 'output-modified': {
                        console.log(`ℹ️${prefix} Output files were modified since the previous run.`);
                        break;
                    }
                    case 'watch-run-start': {
                        if (process.stdout.isTTY) {
                            // If we are in an interactive terminal (TTY), reset it before
                            // each run. This is helpful because it means only the output for
                            // the current build is visible. This is exactly the same as what
                            // "tsc --watch" does.
                            //
                            // This string is the ESC character (ASCII \x1B) followed by "c",
                            // which is the VT100 reset sequence, supported by most terminals:
                            // https://www2.ccs.neu.edu/research/gpc/VonaUtils/vona/terminal/vtansi.htm#:~:text=Reset%20Device
                            console.log('\x1Bc');
                        }
                        break;
                    }
                    case 'watch-run-end': {
                        console.log(`👀${prefix} Watching for file changes`);
                        break;
                    }
                    case 'generic': {
                        console.log(`ℹ️${prefix} ${event.message}`);
                        break;
                    }
                }
            }
        }
    }
}
_DefaultLogger_rootPackageDir = new WeakMap(), _DefaultLogger_diagnosticPrinter = new WeakMap(), _DefaultLogger_instances = new WeakSet(), _DefaultLogger_label = function _DefaultLogger_label(script) {
    const packageDir = script.packageDir;
    const scriptName = 'name' in script ? script.name : undefined;
    if (packageDir !== __classPrivateFieldGet(this, _DefaultLogger_rootPackageDir, "f")) {
        const relativePackageDir = pathlib
            .relative(__classPrivateFieldGet(this, _DefaultLogger_rootPackageDir, "f"), script.packageDir)
            // Normalize to posix-style forward-slashes as the path separator, even
            // on Windows which usually uses back-slashes. This way labels match the
            // syntax used in the package.json dependency specifiers (which are
            // already posix style).
            .replace(pathlib.sep, pathlib.posix.sep);
        if (scriptName !== undefined) {
            return `${relativePackageDir}:${scriptName}`;
        }
        else {
            return relativePackageDir;
        }
    }
    else if (scriptName !== undefined) {
        return scriptName;
    }
    return '';
};
//# sourceMappingURL=default-logger.js.map