var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _PackageJson_instances, _PackageJson_fileAstNode, _PackageJson_scripts, _PackageJson_getOrMakeScriptInfo, _PackageJson_analyzeScriptsSection, _PackageJson_analyzeWireitSection;
/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
import { findNamedNodeAtLocation } from './ast.js';
import { failUnlessJsonObject, failUnlessNonBlankString } from '../analyzer.js';
import { offsetInsideNamedNode, offsetInsideRange } from '../error.js';
/**
 * A parsed and minimally analyzed package.json file.
 *
 * This does some very basic syntactic analysis of the package.json file,
 * finding issues like "the scripts section isn't an object mapping strings to
 * strings" and "the wireit section isn't an object mapping strings to objects".
 *
 * Makes it easy to find the syntax nodes for a script.
 *
 * Does not do any validation or analysis of the wireit script configs.
 *
 * This class exists in part so that we walk the package.json file only once,
 * and in part so that we generate file-level syntactic diagnostics only once,
 * so that we can do better deduplication of errors.
 */
export class PackageJson {
    constructor(jsonFile, fileAstNode) {
        _PackageJson_instances.add(this);
        Object.defineProperty(this, "jsonFile", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        // We keep the file level AST node private to represent the invariant that
        // we only walk the file once, in this class, and nowhere else.
        _PackageJson_fileAstNode.set(this, void 0);
        _PackageJson_scripts.set(this, new Map());
        Object.defineProperty(this, "failures", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "scriptsSection", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: undefined
        });
        Object.defineProperty(this, "wireitSection", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: undefined
        });
        this.jsonFile = jsonFile;
        __classPrivateFieldSet(this, _PackageJson_fileAstNode, fileAstNode, "f");
        const failures = [];
        this.scriptsSection = __classPrivateFieldGet(this, _PackageJson_instances, "m", _PackageJson_analyzeScriptsSection).call(this, failures);
        this.wireitSection = __classPrivateFieldGet(this, _PackageJson_instances, "m", _PackageJson_analyzeWireitSection).call(this, failures);
        this.failures = failures;
    }
    getScriptInfo(name) {
        return __classPrivateFieldGet(this, _PackageJson_scripts, "f").get(name);
    }
    get scripts() {
        return __classPrivateFieldGet(this, _PackageJson_scripts, "f").values();
    }
    getInfoAboutLocation(offset) {
        if (this.scriptsSection && offsetInsideRange(offset, this.scriptsSection)) {
            for (const scriptSyntaxInfo of this.scripts) {
                if (scriptSyntaxInfo.scriptNode &&
                    offsetInsideNamedNode(offset, scriptSyntaxInfo.scriptNode)) {
                    return { kind: 'scripts-section-script', scriptSyntaxInfo };
                }
            }
        }
        else if (this.wireitSection &&
            offsetInsideRange(offset, this.wireitSection)) {
            for (const scriptSyntaxInfo of this.scripts) {
                if (scriptSyntaxInfo.wireitConfigNode &&
                    offsetInsideNamedNode(offset, scriptSyntaxInfo.wireitConfigNode)) {
                    return { kind: 'wireit-section-script', scriptSyntaxInfo };
                }
            }
        }
    }
}
_PackageJson_fileAstNode = new WeakMap(), _PackageJson_scripts = new WeakMap(), _PackageJson_instances = new WeakSet(), _PackageJson_getOrMakeScriptInfo = function _PackageJson_getOrMakeScriptInfo(name) {
    let info = __classPrivateFieldGet(this, _PackageJson_scripts, "f").get(name);
    if (info == null) {
        info = { name };
        __classPrivateFieldGet(this, _PackageJson_scripts, "f").set(name, info);
    }
    return info;
}, _PackageJson_analyzeScriptsSection = function _PackageJson_analyzeScriptsSection(failures) {
    const scriptsSectionResult = findNamedNodeAtLocation(__classPrivateFieldGet(this, _PackageJson_fileAstNode, "f"), ['scripts'], this.jsonFile);
    if (!scriptsSectionResult.ok) {
        failures.push(scriptsSectionResult.error);
        return;
    }
    const scriptsSection = scriptsSectionResult.value;
    if (scriptsSection == null) {
        return;
    }
    const fail = failUnlessJsonObject(scriptsSection, this.jsonFile);
    if (fail != null) {
        failures.push(fail);
        return;
    }
    for (const child of scriptsSection.children ?? []) {
        if (child.type !== 'property') {
            continue;
        }
        const [rawName, rawValue] = child.children ?? [];
        if (rawName == null || rawValue == null) {
            continue;
        }
        const nameResult = failUnlessNonBlankString(rawName, this.jsonFile);
        if (!nameResult.ok) {
            failures.push(nameResult.error);
            continue;
        }
        const valueResult = failUnlessNonBlankString(rawValue, this.jsonFile);
        if (!valueResult.ok) {
            failures.push(valueResult.error);
            continue;
        }
        const scriptAstNode = valueResult.value;
        scriptAstNode.name = nameResult.value;
        __classPrivateFieldGet(this, _PackageJson_instances, "m", _PackageJson_getOrMakeScriptInfo).call(this, nameResult.value.value).scriptNode =
            scriptAstNode;
    }
    return scriptsSectionResult.value;
}, _PackageJson_analyzeWireitSection = function _PackageJson_analyzeWireitSection(failures) {
    const wireitSectionResult = findNamedNodeAtLocation(__classPrivateFieldGet(this, _PackageJson_fileAstNode, "f"), ['wireit'], this.jsonFile);
    if (!wireitSectionResult.ok) {
        failures.push(wireitSectionResult.error);
        return;
    }
    const wireitSection = wireitSectionResult.value;
    if (wireitSection == null) {
        return;
    }
    const fail = failUnlessJsonObject(wireitSection, this.jsonFile);
    if (fail != null) {
        failures.push(fail);
        return;
    }
    for (const child of wireitSection.children ?? []) {
        if (child.type !== 'property') {
            continue;
        }
        const [rawName, rawValue] = child.children ?? [];
        if (rawName == null || rawValue == null) {
            continue;
        }
        const nameResult = failUnlessNonBlankString(rawName, this.jsonFile);
        if (!nameResult.ok) {
            failures.push(nameResult.error);
            continue;
        }
        const fail = failUnlessJsonObject(rawValue, this.jsonFile);
        if (fail != null) {
            failures.push(fail);
            continue;
        }
        const wireitConfigNode = rawValue;
        wireitConfigNode.name = nameResult.value;
        __classPrivateFieldGet(this, _PackageJson_instances, "m", _PackageJson_getOrMakeScriptInfo).call(this, nameResult.value.value).wireitConfigNode =
            wireitConfigNode;
    }
    return wireitSectionResult.value;
};
//# sourceMappingURL=package-json.js.map