/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _WorkerPool_availableWorkers, _WorkerPool_waitingWorkers;
import { Deferred } from './deferred.js';
/**
 * A mechanism for ensuring that at most N tasks are taking place at once.
 *
 * Useful in Wireit to prevent running too many scripts at once and swamping
 * the system. For unlimited parallelism, just set numWorkers to Infinity.
 *
 * Note that node is still single threaded by default. This is useful for
 * Wireit because almost all work is happening in script commands which run
 * in separate processes.
 *
 * No guarantee is made about ordering or fairness of scheduling, though
 * as implemented it's currently LIFO. Deadlocks may occur if there are
 * dependencies between tasks.
 */
export class WorkerPool {
    constructor(numWorkers) {
        _WorkerPool_availableWorkers.set(this, void 0);
        _WorkerPool_waitingWorkers.set(this, []);
        if (numWorkers <= 0) {
            throw new Error(`WorkerPool needs a positive number of workers, got ${numWorkers}`);
        }
        __classPrivateFieldSet(this, _WorkerPool_availableWorkers, numWorkers, "f");
    }
    /**
     * Calls workFn and returns its result.
     *
     * However, no more than `numWorkers` simultaneous calls to workFns will
     * be running at any given time, to prevent overloading the machine.
     */
    async run(workFn) {
        var _a, _b;
        if (__classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f") <= 0) {
            const waiter = new Deferred();
            __classPrivateFieldGet(this, _WorkerPool_waitingWorkers, "f").push(waiter);
            await waiter.promise;
            if (__classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f") <= 0) {
                throw new Error(`Internal error: expected availableWorkers to be positive after task was awoken, but was ${__classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f")}`);
            }
        }
        __classPrivateFieldSet(this, _WorkerPool_availableWorkers, (_a = __classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f"), _a--, _a), "f");
        try {
            return await workFn();
        }
        finally {
            __classPrivateFieldSet(this, _WorkerPool_availableWorkers, (_b = __classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f"), _b++, _b), "f");
            if (__classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f") <= 0) {
                // We intend to override any return or throw with this error in this
                // case.
                // eslint-disable-next-line no-unsafe-finally
                throw new Error(`Internal error: expected availableWorkers to be positive after incrementing, but was ${__classPrivateFieldGet(this, _WorkerPool_availableWorkers, "f")}`);
            }
            __classPrivateFieldGet(this, _WorkerPool_waitingWorkers, "f").pop()?.resolve();
        }
    }
}
_WorkerPool_availableWorkers = new WeakMap(), _WorkerPool_waitingWorkers = new WeakMap();
//# sourceMappingURL=worker-pool.js.map